@@ -25,8 +25,10 @@ from npyx.utils import npa
 #%% IO utilities
 
 def list_files(directory, extension):
-    return [f for f in os.listdir(directory) if f.endswith('.' + extension)]
-
+    files = [f for f in os.listdir(directory) if f.endswith('.' + extension)]
+    files.sort()
+    return files
+    
 #%% Extract metadata and sync channel
 
 def read_spikeglx_meta(dp, subtype='ap'):
@@ -96,10 +98,10 @@ def chan_map(dp=None, y_orig='surface', probe_version=None):
     
     if probe_version in ['3A', '1.0_staggered']:
         Nchan=384
-        cm_el = npa([[  27,   0],
-                           [  59,   0],
-                           [  11,   20],
-                           [  43,   20]])
+        cm_el = npa([[  43,   0],
+                           [  11,   0],
+                           [  59,   20],
+                           [  27,   20]])
         vert=npa([[  0,   40],
                   [  0,   40],
                   [  0,   40],
@@ -112,8 +114,8 @@ def chan_map(dp=None, y_orig='surface', probe_version=None):
     
     elif probe_version=='1.0_aligned':
         Nchan=384
-        cm_el = npa([[  11,   0],
-                     [  43,   0]])
+        cm_el = npa([[  43,   0],
+                     [  11,   0]])
         vert=npa([[  0,   20],
                   [  0,   20]])
         
@@ -124,8 +126,8 @@ def chan_map(dp=None, y_orig='surface', probe_version=None):
         
     elif probe_version=='2.0_singleshank':
         Nchan=384
-        cm_el = npa([[  0,   0],
-                           [  32,   0]])
+        cm_el = npa([[  32,   0],
+                     [  0,   0]])
         vert=npa([[  0,   15],
                   [  0,   15]])
         
@@ -198,7 +200,7 @@ def get_npix_sync(dp, output_binary = False, sourcefile='ap', unit='seconds'):
                     onsets[file_i]=np.load(Path(sync_dp,file))/srate
                     offsets[file_i]=np.load(Path(sync_dp,file[:-13]+'f'+file[-12:]))/srate
                         
-                    return onsets, offsets
+        return onsets, offsets
 
     # Tries to load pre-saved compressed binary
     if op.exists(sync_dp):
@@ -591,7 +593,7 @@ def paq_read(file_path=None):
         num_chars = int(np.fromfile(fid, dtype='>f', count=1))
         chan_name = ''
         for j in range(num_chars):
-            chan_name = chan_name + chr(np.fromfile(fid, dtype='>f', count=1))
+            chan_name = chan_name + chr(int(np.fromfile(fid, dtype='>f', count=1)))
         chan_names.append(chan_name)
 
     # get channel hardware lines
@@ -600,7 +602,7 @@ def paq_read(file_path=None):
         num_chars = int(np.fromfile(fid, dtype='>f', count=1))
         hw_chan = ''
         for j in range(num_chars):
-            hw_chan = hw_chan + chr(np.fromfile(fid, dtype='>f', count=1))
+            hw_chan = hw_chan + chr(int(np.fromfile(fid, dtype='>f', count=1)))
         hw_chans.append(hw_chan)
 
     # get acquisition units
@@ -609,7 +611,7 @@ def paq_read(file_path=None):
         num_chars = int(np.fromfile(fid, dtype='>f', count=1))
         unit = ''
         for j in range(num_chars):
-            unit = unit + chr(np.fromfile(fid, dtype='>f', count=1))
+            unit = unit + chr(int(np.fromfile(fid, dtype='>f', count=1)))
         units.append(unit)
 
     # get data
@@ -810,4 +812,4 @@ def _range_from_slice(myslice, start=None, stop=None, step=None, length=None):
     return myrange
 
 
-from npyx.gl import get_rec_len, assert_multi, get_ds_table
\ No newline at end of file
+from npyx.gl import get_rec_len, assert_multi, get_ds_table
